#include <stddef.h> /* for NULL */
#include <errno.h>
#include <string.h> /* for strerror() */
#include <unistd.h> /* for getlogin() */
#include "config.h" /* for PACKAGE */
#include "support.h"
#include "dialog.h"

const char * config_names[] = 
{
	"url_prefix",
	"thumbnail_prefix",
	"upload_protocol",
	"upload_user",
	"upload_password",
	"upload_host",
	"upload_directory",
	"img_max_width",
	"img_max_height",
	"img_max_is_absolute",
	"img_preferred_format",
	"tn_max_width",
	"tn_max_height",
	"tn_max_is_absolute",
	"tn_preferred_format",
	"preview_update_interval_ms",
	NULL
};

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <glib.h>
#include "config_file.h"

const char * config_filename = "." PACKAGE;
char * config[C_END];
const char * config_filename;
const char * config_filename_tmp;


int init_config(void)
{
	GString * str;
	char *home;

	/* Empty config */
	memset(config, 0, sizeof(config));
	
	/* Fill in config file path */
	home = getenv("HOME");
	if (home == NULL)
		return -1;

	str = g_string_sized_new(128);
	g_string_printf(str, "%s%c%s", home, '/', config_filename);
	config_filename = strdup(str->str);
	g_string_append(str, "-XXXXXX");
	config_filename_tmp = str->str;
	g_string_free(str, FALSE);

	return 0;
}

enum config_variable str2var(const char * str)
{
	int var;
	for (var = 0; var < C_END; var++)
		if (strcmp(config_names[var], str) == 0)
				return var;
	return C_END;
}

int free_config(char * config[])
{
	enum config_variable var;
	for (var = 0; var < C_END; var++)
		if (config[var])
		{
			free(config[var]);
			config[var] = NULL;
		}
	return 0;
}

void set_config(char * dest[], char * src[])
{
	enum config_variable var;
	for (var = 0; var < C_END; var++)
	{
		if (dest[var])
			free(dest[var]);
		dest[var] = src[var];
	}	
}

int set_option_if_unset(char *config[], int i, char * val)
{
	if (val != NULL)
		if ((config[i] == NULL) || (config[i][0] == '\0'))
		{
			config[i] = strdup(val);
			return 1;
		}
	return 0;
}

int set_config_defaults(char *config[])
{
	int changed = 0;
	changed += set_option_if_unset(config,
			C_URL_PREFIX, "http://your.host.com/subdir");
	changed += set_option_if_unset(config,
			C_THUMB_PREFIX, "tn-");
	changed += set_option_if_unset(config,
			C_UPLOAD_USER, getlogin());
	changed += set_option_if_unset(config,
			C_PREVIEW_UPD_INT, "0.5");

	return changed;
}

#define CHK(cond) \
	if (cond) \
	{ \
		if (filename != NULL) \
			fprintf(stderr, "Invalid string at %s:%d - %s\n", \
				filename, lineno, buf); \
		return 0; \
	}

int parse_config_line(const char *filename, int lineno, char *buf,
		char ** rname, char ** rval)
{
	char * name;
	int namelen;
	char * value;
	int valuelen;
	char * ptr;

	if (buf[0] == '#')
		return 0; /* comment */
	/* remove '\n' */
	buf[strlen(buf) - 1] = '\0';
	name = buf; namelen = 0;
	while (isspace(*name) && (*name != '\0'))
		name++;
	if (*name == '\0')
		return 0; /* empty line */
	while ((isalnum(name[namelen]) || (name[namelen] == '_'))
			&& (*name != '\0'))
		namelen++;
	value = name + namelen;
	CHK(value == '\0' )
	while (isspace(*value) && (*value != '\0'))
		value++;
	CHK(*value != '=')
	value++;
	while (isspace(*value) && (*value != '\0'))
		value++;
	valuelen = 0;
	while (isprint(value[valuelen]) && (value[valuelen] != '\0'))
		valuelen++;
/*	CHK(valuelen == 0) */
	ptr = value + valuelen;
	while (isspace(*ptr) && (*ptr != '\0'))
		ptr++;
	CHK(*ptr != '\0')

/*	printf("BUF %s\n", buf); */
	name[namelen] = '\0';
	value[valuelen] = '\0';
/*	printf("NAME '%s' (len %d) VALUE '%s' (len %d)\n\n",
			name, namelen, value, valuelen);
*/
	*rname = name;
	*rval = value;
	return 1;
}

int read_config(char * config[])
{
	const char * fn;
	FILE * fd;
	char buf[BUFSIZ];
	int lineno;
	char * name;
	char * value;
	enum config_variable var;

	fn = config_filename;
	fd = fopen(fn, "r");
	if (fd == NULL)
		return -1;

	lineno = 0;
	while (!feof(fd))
	{
		if (fgets(buf, sizeof(buf), fd) == NULL)
			break;
		lineno++;
		if (parse_config_line(fn, lineno, buf, &name, &value) == 0)
			continue;
		var = str2var(name);
		if (var == C_END)
		{
			fprintf(stderr, "Unknown variable \"%s\" at %s:%d\n",
					name, fn, lineno);
			continue;
		}
/*		printf("[line %d] %s = '%s'\n", lineno,
				config_names[var], value);
*/
		if (config[var])
		{
			fprintf(stderr, "Warning: duplicate entry for "
					"parameter '%s' at %s:%d (ignored)\n",
					name, fn, lineno);
			continue;
		}
		config[var] = strdup(value);
	}
	return 0;
}

int write_config(char * config[])
{
	FILE * fdr = NULL;
	FILE * fdw;
	char buf[BUFSIZ];
	char buf_orig[BUFSIZ];
	int lineno;
	int is_written[C_END];
	char * name;
	char * value;
	enum config_variable var;
	char * fnw;
	int fd;
	int ret;

	fnw = strdup(config_filename_tmp);
	fd = mkstemp(fnw);
	if (fd < 0)
		return -1;
	fdw = fdopen(fd, "w");
	if (fdw == NULL)
		return -1;
	
	memset(is_written, 0, sizeof(is_written));
	
	fdr = fopen(config_filename, "r");
	if (fdr == NULL)
		/* There's no config file - let's write one */
		goto write_tail;
	
	lineno = 0;
	while (!feof(fdr))
	{
		if (fgets(buf, sizeof(buf), fdr) == NULL)
			break; /* EOF reached */
		lineno++;
		strcpy(buf_orig, buf);
		if (parse_config_line(NULL, lineno, buf, &name, &value) == 0)
			/* error or empty line - just write it as is */
			goto writeline;
		var = str2var(name);
		if (var == C_END)
			/* unknown variable - just write as is */
			goto writeline;
		/* Known variable found, let's see */
		if (is_written[var] == 1) /* already written */
			goto writeline;
		/* We will write it past this point */
		is_written[var] = 1;

		if (config[var] == NULL)
		{
			/* variable is not set */
			goto writeline;
		}
		/* Check if var is changed */
		if (strcmp(value, config[var]) != 0)
		{
			/* Put the new value into line */
			strcpy(buf_orig + (value - buf), config[var]);
			strcat(buf_orig + (value - buf), "\n");
		}
		
writeline:
		fputs(buf_orig, fdw);
	}

write_tail:
	/* Write those variables that were not written yet */
	for (var = 0; var < C_END; var++)
	{
		if ((! is_written[var]) &&
				(config[var] != NULL) &&
				(config[var][0] != '\0') &&
				(config[var][0] != '0'))
		{
			fprintf(fdw, "%s = %s\n",
				config_names[var], config[var]);
		}
	}

	/* Rename new file to old */
	if (rename(fnw, config_filename) != 0)
	{
		GString * s = g_string_sized_new (128);
		g_string_printf (s, _("rename (\"%s\", \"%s\") error:\n%s"),
				fnw, config_filename, strerror(errno));
		show_error_tt (_("Error saving configuration file"),
				_("<big><b>Error:</b> can't save "
					"configuration file.</big>"),
				s->str, s->len);
		g_string_free (s, TRUE);
		ret = -1;
		goto out;
	}

	/* Everything went fine...oh wow! */
	ret = 0;
out:
	free(fnw);
	if (fdr != NULL)
		fclose(fdr);
	fclose(fdw);
	return ret;
}

#ifdef TEST_CONFIG
int main(void)
{
	char * config[C_END];

	if (init_config_filename() != 0)
	{
		fprintf(stderr, "ERROR: $HOME is not set - exiting\n");
		return 1;
	}
	
	read_config(config);

	config[C_URL_PREFIX] = "http://kir.vtx.ru/lj/";
	config[C_UPLOAD_HOST] = "kir.vtx.ru";

	return write_config(config);
}
#endif
