/* upload.c
 * 
 * Copyright (C) 2003 by Kir Kolyshkin <kir@sacred.ru>
 * 
 * Functions that implements file upload via ftp (libcurl is used) or scp.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h> /* for free() */
#include <errno.h>
#include <curl/curl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>
#include "support.h"
#include "upload.h"
#include "dialog.h"
#include "filename.h"
#include "config_file.h"

const char * upload_cmd_scp = "setsid scp %f %u@%h:%d 2>&1";

int upload_files_curl(const char **files,
		const char **config);

int upload_files_extcmd(const char **files,
		const char **config,
		const char * str);

void show_upload_error(const char * output, int len)
{
	show_error_tt (_("Error uploading image"),
			_("<big><b>Error:</b> upload failed</big>"),
			output, len);
}

/* Available options:
 *  %f - list of files, separated by spaces
 *  %p - protocol (e.g. ftp)
 *  %u - user name
 *  %w - password
 *  %h - host name
 *  %d - path (directory)
 *  %% - per cent sign (%)
 */
GString * parse_upload_str(const char ** files,
		const char ** config,
		const char *str)
{
	GString * buf;
	const char * cmd = str;

	buf = g_string_sized_new (128);
	do
	{
		if (*cmd != '%')
		{
			g_string_append_c(buf, *cmd);
			continue;
		}
		/* parse special symbol */
		cmd++;
		switch (*cmd)
		{
			case 'f':
				while (*files)
				{
					g_string_append(buf, *files++);
					if (*files)
						g_string_append_c(buf, ' ');
				}
				break;
			case 'p':
				g_string_append(buf, config[C_UPLOAD_PROTO]);
				break;
			case 'u':
				g_string_append(buf, config[C_UPLOAD_USER]);
				break;
			case 'w':
				g_string_append(buf, config[C_UPLOAD_PASS]);
				break;
			case 'h':
				g_string_append(buf, config[C_UPLOAD_HOST]);
				break;
			case 'd':
				g_string_append(buf, config[C_UPLOAD_PATH]);
				break;
			case '%':
				g_string_append_c(buf, '%');
				break;
			default:
				fprintf(stderr, "WARNING: unknown option %%%c "
						"in 'upload_cmd'\n", *cmd);
				break;
		}
	} while (*(++cmd) != '\0');
	
	return buf;
}

int upload_files(const char ** files, const char ** config)
{
	if (strcmp(config[C_UPLOAD_PROTO], "ftp") == 0)
	{
		return upload_files_curl(files, config);
	}
	else 
	{
		if (strcmp(config[C_UPLOAD_PROTO], "scp") == 0)
		{
			return upload_files_extcmd(files, config,
				upload_cmd_scp);
		}
		else
		{
			return -1;
		}
	}
}

int upload_files_extcmd(const char *files[],
		const char ** config,
		const char * str)
{
	FILE *f;
	GString * cmd;
	char buf[BUFSIZ];
	int ret;
	GString * output;

	cmd = parse_upload_str(files, config, str);
	output = g_string_sized_new (BUFSIZ);

	fprintf (stderr, "+ %s\n", cmd->str);
	f = popen (cmd->str, "r");
	if (f == NULL)
	{
		g_string_printf(output, _("popen(\"%s\") failed: %s"),
				cmd->str, strerror(errno));
		show_upload_error (output->str, output->len);
		return -1;
	}	
	while (fgets(buf, BUFSIZ, f) != NULL)
		g_string_append(output, buf);
	ret = pclose(f);
	if (ret != 0)
		show_upload_error (output->str, output->len);
/*
	printf("-- Command output follows: --\n%s\n--- END ---\n",
		output->str);
*/
	g_string_free (output, TRUE);
	g_string_free (cmd, TRUE);
	return (ret == 0)? 0: -2;
}

int upload_files_curl(const char **files,
		const char ** config)
{
	CURL *curl;
	CURLcode res;
	struct stat file_info;
	FILE * f;
	char errorbuf[CURL_ERROR_SIZE];
	GString * remote_url;
	const char * onefile[2];
	const char * filename;

	onefile[1] = NULL;
	curl_global_init(CURL_GLOBAL_ALL);
	curl = curl_easy_init();
	if (curl == NULL)
	{
		show_upload_error(_("Can't initialize curl: "
					"curl_easy_init() failed"), -1);
		curl_global_cleanup();
		return -2;
	}
	curl_easy_setopt(curl, CURLOPT_UPLOAD, TRUE);
	curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, errorbuf);
	res = -1;
	while ((filename = *files++) != NULL)
	{
		f = fopen(filename, "r");
		if (f == NULL)
		{
			GString * err;
			/* display can't open file error */
			err = g_string_sized_new (128);
			g_string_printf(err,
				_("Error opening file %s: %s"),
				filename, strerror(errno));
			show_upload_error(err->str, err->len);
			g_string_free (err, TRUE);

			goto cleanup;
		}
		fstat(fileno(f), &file_info);
		onefile[0] = getfilename(filename);
		remote_url = parse_upload_str(onefile, config,
				"%p://%u:%w@%h/%d/%f");
/*		printf("cURL = %s\n", remote_url->str); */
		curl_easy_setopt(curl, CURLOPT_URL, remote_url->str);
		curl_easy_setopt(curl, CURLOPT_INFILE, f);
		curl_easy_setopt(curl, CURLOPT_INFILESIZE, file_info.st_size);
		res = curl_easy_perform(curl);
		g_string_free (remote_url, TRUE);
		if (res != 0)
		{
			show_upload_error(errorbuf, -1);
			goto cleanup;
		}
	}
cleanup:
	curl_easy_cleanup(curl);
	curl_global_cleanup();
	return res;
}
